/*
 *  Copyright (C) 2008 dhewg, #wiidev efnet
 *
 *  this file is part of geckoloader
 *  http://wiibrew.org/index.php?title=Geckoloader
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <ogcsys.h>

#include "usb.h"
#include "flash.h"
#include "delay.h"
#include "../../config.h"

extern void __exception_closeall ();

extern u8 stub_bin[];
extern u32 stub_bin_size;
extern u8 loader_dol[];
extern u32 loader_dol_size;

static void *xfb = NULL;
static GXRModeObj *rmode;

void (*reload) () = (void (*) ()) STUB_MEM_ADDR;

void gecko_write_byte (unsigned long address, const const u8 value) {
        u8 i, c;

        c = ~value;
        for (i = 0; i < 5; ++i) {
                flash_programbyte (address, value);
                c = flash_readcommand (address);

                if (c == value)
                        return;

                printf ("error writing to address 0x%08lx, retry #%u\n",
                        address, i + 1);

                mdelay (250);
        }

        if (c != value) {
                printf ("fatal error writing to the usbgecko flash!\n"
                        "if this problem occures again at the same offset,\n"
                        "try setting a different one in config.h\n\n"
                        "halted.\n");

                while (1)
                        VIDEO_WaitVSync ();
        }
}

void gecko_write_flash () {
        u32 i;

        usb_flush ();

        printf ("erasing the usbgecko flash...\n\n");

        flash_chiperase ();

        mdelay (3000);

        printf ("writing the loader code to usbgecko flash\n"
                "(0x%08x bytes at offset 0x%08x)\n\n"
                "please wait...\n",
                loader_dol_size, FLASH_OFFSET);

        gecko_write_byte (FLASH_OFFSET + 0, (loader_dol_size >> 24) & 0xff);
        gecko_write_byte (FLASH_OFFSET + 1, (loader_dol_size >> 16) & 0xff);
        gecko_write_byte (FLASH_OFFSET + 2, (loader_dol_size >> 8) & 0xff);
        gecko_write_byte (FLASH_OFFSET + 3, loader_dol_size & 0xff);

        for (i = 0; i < loader_dol_size; ++i)
                gecko_write_byte (FLASH_OFFSET + 4 + i, loader_dol[i]);

        printf ("success!\n\n");
        mdelay (3000);
}

int main() {
        VIDEO_Init ();

        switch (VIDEO_GetCurrentTvMode())
        {
                case VI_NTSC:
                        rmode = &TVNtsc480IntDf;
                        break;
                case VI_PAL:
                        rmode = &TVPal528IntDf;
                        break;
                case VI_MPAL:
                        rmode = &TVMpal480IntDf;
                        break;
                default:
                        rmode = &TVNtsc480IntDf;
                        break;
        }

        PAD_Init ();

        xfb = MEM_K0_TO_K1 (SYS_AllocateFramebuffer (rmode));

        VIDEO_Configure (rmode);
        VIDEO_SetNextFramebuffer (xfb);
        VIDEO_SetBlack (FALSE);
        VIDEO_Flush ();
        VIDEO_WaitVSync ();
        if (rmode->viTVMode&VI_NON_INTERLACE)
                VIDEO_WaitVSync ();

        console_init(xfb, 20, 64, rmode->fbWidth, rmode->xfbHeight,
                     rmode->fbWidth * 2);
        VIDEO_SetNextFramebuffer (xfb);

        printf ("geckoloader boot code " GECKOLOADER_VERSION
#ifdef DEBUG_GECKOLOADER
                " DEBUG VERSION\n"
#else
                "\n"
#endif
                "coded by dhewg, #wiidev efnet/blitzed\n\n"
                "stub size: 0x%04x loader size: 0x%08x\n\n"
                "press A to boot a prior flashed loader\n"
                "press Z to write the loader code to the usbgecko flash\n\n",
                stub_bin_size, loader_dol_size);

        while (1) {
                PAD_ScanPads ();
                if (PAD_ButtonsDown (0) & PAD_BUTTON_A ||
                    PAD_ButtonsDown (0) & PAD_TRIGGER_Z)
                        break;
                VIDEO_WaitVSync ();
        }

        if (PAD_ButtonsDown (0) & PAD_TRIGGER_Z)
                gecko_write_flash ();
        
        printf ("installing the stub (0x%x bytes to 0x%08x)\n",
                stub_bin_size, STUB_MEM_ADDR);
        VIDEO_WaitVSync ();

        memcpy ((void *) STUB_MEM_ADDR, stub_bin, stub_bin_size);
        DCFlushRange ((void *) STUB_MEM_ADDR, stub_bin_size);

        printf ("executing the stub\n");
        VIDEO_WaitVSync ();

        reload ();

        printf ("eh?\n");
        VIDEO_WaitVSync ();

        return 0;
}

