/*
 *  Copyright (C) 2008 dhewg, #wiidev efnet
 *
 *  this file is part of geckoloader
 *  http://wiibrew.org/index.php?title=Geckoloader
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "gecko.h"
#include "../config.h"

unsigned char cmd_senddol = 0x14;
const char *envvar = "USBGECKODEVICE";

void wait_for_ack () {
        unsigned char ack;
        if (gecko_read (&ack, 1))
                fprintf (stderr, "error receiving the ack\n");
        else if (ack != 0xaa)
                fprintf (stderr, "unknown ack (0x%02x)\n", ack);
}

int main (int argc, char **argv) {
        int fd;
        struct stat stat;
        char *tty;
        unsigned char size[4];;
        unsigned char *buf, *p;
        off_t fsize, block;

        if (argc < 2) {
                fprintf (stderr, "geckoupload " GECKOLOADER_VERSION "\n"
                         "coded by dhewg, #wiidev efnet/blitzed\n\n"
                         "usage: %s <dol file>\n\n", argv[0]);
                exit (EXIT_FAILURE);
        }

        tty = getenv (envvar);
        if (!tty) {
                fprintf (stderr, "please set the environment variable "
                         "%s first (eg \"/dev/ttyUSB0\")\n", envvar);
                exit (EXIT_FAILURE);
        }

        fd = open (argv[1], O_RDONLY);
        if (fd < 0) {
                perror ("error opening the file");
                exit (EXIT_FAILURE);
        }

        if (fstat (fd, &stat)) {
                close (fd);
                perror ("error stat'ing the file");
                exit (EXIT_FAILURE);
        }
        fsize = stat.st_size;

        if (fsize < 1 || fsize > 8 * 1024 * 1024) {
                close (fd);
                fprintf (stderr, "error: invalid file size\n");
                exit (EXIT_FAILURE);
        }

        buf = malloc (fsize);
        if (!buf) {
                close (fd);
                fprintf (stderr, "out of memory\n");
                exit (EXIT_FAILURE);
        }

        if (read (fd, buf, fsize) != fsize) {
                close (fd);
                free (buf);
                perror ("error reading the file");
                exit (EXIT_FAILURE);
        }
        close (fd);

        if (gecko_open (tty)) {
                free (buf);
                fprintf (stderr, "unable to open the device\n");
                exit (EXIT_FAILURE);
        }

        printf ("sending upload request\n");
        if (gecko_write (&cmd_senddol, 1)) {
                free (buf);
                gecko_close ();
                exit (EXIT_FAILURE);
        }

        printf ("awaiting upload ack\n");
        wait_for_ack ();

        size[0] = (fsize >> 24) & 0xff;
        size[1] = (fsize >> 16) & 0xff;
        size[2] = (fsize >> 8) & 0xff;
        size[3] = fsize & 0xff;

        printf ("sending file size (0x%02x%02x%02x%02x)\n",
                size[0], size[1], size[2], size[3]);

        if (gecko_write (size, 4)) {
                free (buf);
                gecko_close ();
                fprintf (stderr, "error sending data\n");
                exit (EXIT_FAILURE);
        }

        printf ("sending data");
        fflush (stdout);

        p = buf;
        while (fsize > 0) {
                block = fsize;
                if (block > 63488)
                        block = 63488;
                fsize -= block;

                if (gecko_write (p, block)) {
                        fprintf (stderr, "error sending block\n");
                        break;
                }
                p += block;

                printf (".");
                fflush (stdout);

                if (fsize > 0)
                        wait_for_ack ();
                else
                        printf ("\n");
        }

        printf ("done.\n");

        free (buf);
        gecko_close ();

        return 0;
}

